package org.eclipse.jdt.internal.compiler.lookup;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.List;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ClassFile;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Argument;
import org.eclipse.jdt.internal.compiler.ast.LambdaExpression;
import org.eclipse.jdt.internal.compiler.ast.MethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.codegen.ConstantPool;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationHolder;
import org.eclipse.jdt.internal.compiler.lookup.ArrayBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.InvocationSite;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodVerifier;
import org.eclipse.jdt.internal.compiler.lookup.PackageBinding;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedGenericMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.Substitution;
import org.eclipse.jdt.internal.compiler.lookup.SyntheticArgumentBinding;
import org.eclipse.jdt.internal.compiler.lookup.SyntheticMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.util.Util;

public class MethodBinding extends Binding {
  public int modifiers;
  
  public char[] selector;
  
  public TypeBinding returnType;
  
  public TypeBinding[] parameters;
  
  public TypeBinding receiver;
  
  public ReferenceBinding[] thrownExceptions;
  
  public ReferenceBinding declaringClass;
  
  public TypeVariableBinding[] typeVariables = Binding.NO_TYPE_VARIABLES;
  
  char[] signature;
  
  public long tagBits;
  
  protected AnnotationBinding[] typeAnnotations = Binding.NO_ANNOTATIONS;
  
  public Boolean[] parameterNonNullness;
  
  public int defaultNullness;
  
  public char[][] parameterNames = Binding.NO_PARAMETER_NAMES;
  
  protected MethodBinding() {}
  
  public MethodBinding(int modifiers, char[] selector, TypeBinding returnType, TypeBinding[] parameters, ReferenceBinding[] thrownExceptions, ReferenceBinding declaringClass) {
    this.modifiers = modifiers;
    this.selector = selector;
    this.returnType = returnType;
    this.parameters = (parameters == null || parameters.length == 0) ? Binding.NO_PARAMETERS : parameters;
    this.thrownExceptions = (thrownExceptions == null || thrownExceptions.length == 0) ? Binding.NO_EXCEPTIONS : thrownExceptions;
    this.declaringClass = declaringClass;
    if (this.declaringClass != null && 
      this.declaringClass.isStrictfp() && 
      !isNative() && !isAbstract())
      this.modifiers |= 0x800; 
  }
  
  public MethodBinding(int modifiers, TypeBinding[] parameters, ReferenceBinding[] thrownExceptions, ReferenceBinding declaringClass) {
    this(modifiers, TypeConstants.INIT, (TypeBinding)TypeBinding.VOID, parameters, thrownExceptions, declaringClass);
  }
  
  public MethodBinding(org.eclipse.jdt.internal.compiler.lookup.MethodBinding initialMethodBinding, ReferenceBinding declaringClass) {
    this.modifiers = initialMethodBinding.modifiers;
    this.selector = initialMethodBinding.selector;
    this.returnType = initialMethodBinding.returnType;
    this.parameters = initialMethodBinding.parameters;
    this.thrownExceptions = initialMethodBinding.thrownExceptions;
    this.declaringClass = declaringClass;
    declaringClass.storeAnnotationHolder(this, initialMethodBinding.declaringClass.retrieveAnnotationHolder(initialMethodBinding, true));
  }
  
  public final boolean areParameterErasuresEqual(org.eclipse.jdt.internal.compiler.lookup.MethodBinding method) {
    TypeBinding[] args = method.parameters;
    if (this.parameters == args)
      return true; 
    int length = this.parameters.length;
    if (length != args.length)
      return false; 
    for (int i = 0; i < length; i++) {
      if (TypeBinding.notEquals(this.parameters[i], args[i]) && TypeBinding.notEquals(this.parameters[i].erasure(), args[i].erasure()))
        return false; 
    } 
    return true;
  }
  
  public final boolean areParametersCompatibleWith(TypeBinding[] arguments) {
    int paramLength = this.parameters.length;
    int argLength = arguments.length;
    int lastIndex = argLength;
    if (isVarargs()) {
      lastIndex = paramLength - 1;
      if (paramLength == argLength) {
        TypeBinding varArgType = this.parameters[lastIndex];
        TypeBinding lastArgument = arguments[lastIndex];
        if (TypeBinding.notEquals(varArgType, lastArgument) && !lastArgument.isCompatibleWith(varArgType))
          return false; 
      } else if (paramLength < argLength) {
        TypeBinding varArgType = ((ArrayBinding)this.parameters[lastIndex]).elementsType();
        for (int j = lastIndex; j < argLength; j++) {
          if (TypeBinding.notEquals(varArgType, arguments[j]) && !arguments[j].isCompatibleWith(varArgType))
            return false; 
        } 
      } else if (lastIndex != argLength) {
        return false;
      } 
    } 
    for (int i = 0; i < lastIndex; i++) {
      if (TypeBinding.notEquals(this.parameters[i], arguments[i]) && !arguments[i].isCompatibleWith(this.parameters[i]))
        return false; 
    } 
    return true;
  }
  
  public final boolean areParametersEqual(org.eclipse.jdt.internal.compiler.lookup.MethodBinding method) {
    TypeBinding[] args = method.parameters;
    if (this.parameters == args)
      return true; 
    int length = this.parameters.length;
    if (length != args.length)
      return false; 
    for (int i = 0; i < length; i++) {
      if (TypeBinding.notEquals(this.parameters[i], args[i]))
        return false; 
    } 
    return true;
  }
  
  public final boolean areTypeVariableErasuresEqual(org.eclipse.jdt.internal.compiler.lookup.MethodBinding method) {
    TypeVariableBinding[] vars = method.typeVariables;
    if (this.typeVariables == vars)
      return true; 
    int length = this.typeVariables.length;
    if (length != vars.length)
      return false; 
    for (int i = 0; i < length; i++) {
      if (TypeBinding.notEquals((TypeBinding)this.typeVariables[i], (TypeBinding)vars[i]) && TypeBinding.notEquals(this.typeVariables[i].erasure(), vars[i].erasure()))
        return false; 
    } 
    return true;
  }
  
  public org.eclipse.jdt.internal.compiler.lookup.MethodBinding asRawMethod(LookupEnvironment env) {
    if (this.typeVariables == Binding.NO_TYPE_VARIABLES)
      return this; 
    int length = this.typeVariables.length;
    TypeBinding[] arguments = new TypeBinding[length];
    for (int i = 0; i < length; i++) {
      TypeVariableBinding var = this.typeVariables[i];
      if (var.boundsCount() <= 1) {
        arguments[i] = env.convertToRawType(var.upperBound(), false);
      } else {
        ReferenceBinding[] arrayOfReferenceBinding = var.superInterfaces();
        int superLength = arrayOfReferenceBinding.length;
        TypeBinding rawFirstBound = null;
        TypeBinding[] rawOtherBounds = null;
        if (var.boundsCount() == superLength) {
          rawFirstBound = env.convertToRawType((TypeBinding)arrayOfReferenceBinding[0], false);
          rawOtherBounds = new TypeBinding[superLength - 1];
          for (int s = 1; s < superLength; s++)
            rawOtherBounds[s - 1] = env.convertToRawType((TypeBinding)arrayOfReferenceBinding[s], false); 
        } else {
          rawFirstBound = env.convertToRawType((TypeBinding)var.superclass(), false);
          rawOtherBounds = new TypeBinding[superLength];
          for (int s = 0; s < superLength; s++)
            rawOtherBounds[s] = env.convertToRawType((TypeBinding)arrayOfReferenceBinding[s], false); 
        } 
        arguments[i] = (TypeBinding)env.createWildcard(null, 0, rawFirstBound, rawOtherBounds, 1);
      } 
    } 
    return (org.eclipse.jdt.internal.compiler.lookup.MethodBinding)env.createParameterizedGenericMethod(this, arguments);
  }
  
  public final boolean canBeSeenBy(InvocationSite invocationSite, Scope scope) {
    if (isPublic())
      return true; 
    SourceTypeBinding invocationType = scope.enclosingSourceType();
    if (TypeBinding.equalsEquals((TypeBinding)invocationType, (TypeBinding)this.declaringClass))
      return true; 
    if (isProtected()) {
      if (invocationType.fPackage == this.declaringClass.fPackage)
        return true; 
      return invocationSite.isSuperAccess();
    } 
    if (isPrivate()) {
      ReferenceBinding referenceBinding1;
      SourceTypeBinding sourceTypeBinding = invocationType;
      ReferenceBinding temp = sourceTypeBinding.enclosingType();
      while (temp != null) {
        referenceBinding1 = temp;
        temp = temp.enclosingType();
      } 
      ReferenceBinding outerDeclaringClass = (ReferenceBinding)this.declaringClass.erasure();
      temp = outerDeclaringClass.enclosingType();
      while (temp != null) {
        outerDeclaringClass = temp;
        temp = temp.enclosingType();
      } 
      return TypeBinding.equalsEquals((TypeBinding)referenceBinding1, (TypeBinding)outerDeclaringClass);
    } 
    return (invocationType.fPackage == this.declaringClass.fPackage);
  }
  
  public final boolean canBeSeenBy(PackageBinding invocationPackage) {
    if (isPublic())
      return true; 
    if (isPrivate())
      return false; 
    return (invocationPackage == this.declaringClass.getPackage());
  }
  
  public final boolean canBeSeenBy(TypeBinding receiverType, InvocationSite invocationSite, Scope scope) {
    SourceTypeBinding invocationType = scope.enclosingSourceType();
    if (this.declaringClass.isInterface() && isStatic()) {
      if ((scope.compilerOptions()).sourceLevel < 3407872L)
        return false; 
      if ((invocationSite.isTypeAccess() || invocationSite.receiverIsImplicitThis()) && TypeBinding.equalsEquals(receiverType, (TypeBinding)this.declaringClass))
        return true; 
      return false;
    } 
    if (isPublic())
      return true; 
    if (TypeBinding.equalsEquals((TypeBinding)invocationType, (TypeBinding)this.declaringClass) && TypeBinding.equalsEquals((TypeBinding)invocationType, receiverType))
      return true; 
    if (invocationType == null)
      return (!isPrivate() && scope.getCurrentPackage() == this.declaringClass.fPackage); 
    if (isProtected()) {
      if (TypeBinding.equalsEquals((TypeBinding)invocationType, (TypeBinding)this.declaringClass))
        return true; 
      if (invocationType.fPackage == this.declaringClass.fPackage)
        return true; 
      SourceTypeBinding sourceTypeBinding = invocationType;
      TypeBinding receiverErasure = receiverType.erasure();
      ReferenceBinding declaringErasure = (ReferenceBinding)this.declaringClass.erasure();
      int depth = 0;
      while (true) {
        if (sourceTypeBinding.findSuperTypeOriginatingFrom((TypeBinding)declaringErasure) != null) {
          if (invocationSite.isSuperAccess())
            return true; 
          if (receiverType instanceof ArrayBinding)
            return false; 
          if (isStatic()) {
            if (depth > 0)
              invocationSite.setDepth(depth); 
            return true;
          } 
          if (TypeBinding.equalsEquals((TypeBinding)sourceTypeBinding, receiverErasure) || receiverErasure.findSuperTypeOriginatingFrom((TypeBinding)sourceTypeBinding) != null) {
            if (depth > 0)
              invocationSite.setDepth(depth); 
            return true;
          } 
        } 
        depth++;
        ReferenceBinding referenceBinding = sourceTypeBinding.enclosingType();
        if (referenceBinding == null)
          return false; 
      } 
    } 
    if (isPrivate()) {
      if (TypeBinding.notEquals(receiverType, (TypeBinding)this.declaringClass))
        if ((scope.compilerOptions()).complianceLevel > 3276800L || !receiverType.isTypeVariable() || !((TypeVariableBinding)receiverType).isErasureBoundTo(this.declaringClass.erasure()))
          return false;  
      if (TypeBinding.notEquals((TypeBinding)invocationType, (TypeBinding)this.declaringClass)) {
        ReferenceBinding referenceBinding1;
        SourceTypeBinding sourceTypeBinding = invocationType;
        ReferenceBinding temp = sourceTypeBinding.enclosingType();
        while (temp != null) {
          referenceBinding1 = temp;
          temp = temp.enclosingType();
        } 
        ReferenceBinding outerDeclaringClass = (ReferenceBinding)this.declaringClass.erasure();
        temp = outerDeclaringClass.enclosingType();
        while (temp != null) {
          outerDeclaringClass = temp;
          temp = temp.enclosingType();
        } 
        if (TypeBinding.notEquals((TypeBinding)referenceBinding1, (TypeBinding)outerDeclaringClass))
          return false; 
      } 
      return true;
    } 
    PackageBinding declaringPackage = this.declaringClass.fPackage;
    if (invocationType.fPackage != declaringPackage)
      return false; 
    if (receiverType instanceof ArrayBinding)
      return false; 
    TypeBinding originalDeclaringClass = this.declaringClass.original();
    ReferenceBinding currentType = (ReferenceBinding)receiverType;
    while (true) {
      if (currentType.isCapture()) {
        if (TypeBinding.equalsEquals(originalDeclaringClass, currentType.erasure().original()))
          return true; 
      } else if (TypeBinding.equalsEquals(originalDeclaringClass, currentType.original())) {
        return true;
      } 
      PackageBinding currentPackage = currentType.fPackage;
      if (!currentType.isCapture() && currentPackage != null && currentPackage != declaringPackage)
        return false; 
      if ((currentType = currentType.superclass()) == null)
        return false; 
    } 
  }
  
  public List<TypeBinding> collectMissingTypes(List<TypeBinding> missingTypes) {
    if ((this.tagBits & 0x80L) != 0L) {
      missingTypes = this.returnType.collectMissingTypes(missingTypes);
      int i, max;
      for (i = 0, max = this.parameters.length; i < max; i++)
        missingTypes = this.parameters[i].collectMissingTypes(missingTypes); 
      for (i = 0, max = this.thrownExceptions.length; i < max; i++)
        missingTypes = this.thrownExceptions[i].collectMissingTypes(missingTypes); 
      for (i = 0, max = this.typeVariables.length; i < max; i++) {
        TypeVariableBinding variable = this.typeVariables[i];
        missingTypes = variable.superclass().collectMissingTypes(missingTypes);
        ReferenceBinding[] interfaces = variable.superInterfaces();
        for (int j = 0, length = interfaces.length; j < length; j++)
          missingTypes = interfaces[j].collectMissingTypes(missingTypes); 
      } 
    } 
    return missingTypes;
  }
  
  org.eclipse.jdt.internal.compiler.lookup.MethodBinding computeSubstitutedMethod(org.eclipse.jdt.internal.compiler.lookup.MethodBinding method, LookupEnvironment env) {
    int length = this.typeVariables.length;
    TypeVariableBinding[] vars = method.typeVariables;
    if (length != vars.length)
      return null; 
    ParameterizedGenericMethodBinding substitute = 
      env.createParameterizedGenericMethod(method, (TypeBinding[])this.typeVariables);
    for (int i = 0; i < length; i++) {
      if (!this.typeVariables[i].isInterchangeableWith(vars[i], (Substitution)substitute))
        return null; 
    } 
    return (org.eclipse.jdt.internal.compiler.lookup.MethodBinding)substitute;
  }
  
  public char[] computeUniqueKey(boolean isLeaf) {
    char[] declaringKey = this.declaringClass.computeUniqueKey(false);
    int declaringLength = declaringKey.length;
    int selectorLength = (this.selector == TypeConstants.INIT) ? 0 : this.selector.length;
    char[] sig = genericSignature();
    boolean isGeneric = (sig != null);
    if (!isGeneric)
      sig = signature(); 
    int signatureLength = sig.length;
    int thrownExceptionsLength = this.thrownExceptions.length;
    int thrownExceptionsSignatureLength = 0;
    char[][] thrownExceptionsSignatures = null;
    boolean addThrownExceptions = (thrownExceptionsLength > 0 && (!isGeneric || CharOperation.lastIndexOf('^', sig) < 0));
    if (addThrownExceptions) {
      thrownExceptionsSignatures = new char[thrownExceptionsLength][];
      for (int i = 0; i < thrownExceptionsLength; i++) {
        if (this.thrownExceptions[i] != null) {
          thrownExceptionsSignatures[i] = this.thrownExceptions[i].signature();
          thrownExceptionsSignatureLength += (thrownExceptionsSignatures[i]).length + 1;
        } 
      } 
    } 
    char[] uniqueKey = new char[declaringLength + 1 + selectorLength + signatureLength + thrownExceptionsSignatureLength];
    int index = 0;
    System.arraycopy(declaringKey, 0, uniqueKey, index, declaringLength);
    index = declaringLength;
    uniqueKey[index++] = '.';
    System.arraycopy(this.selector, 0, uniqueKey, index, selectorLength);
    index += selectorLength;
    System.arraycopy(sig, 0, uniqueKey, index, signatureLength);
    if (thrownExceptionsSignatureLength > 0) {
      index += signatureLength;
      for (int i = 0; i < thrownExceptionsLength; i++) {
        char[] thrownExceptionSignature = thrownExceptionsSignatures[i];
        if (thrownExceptionSignature != null) {
          uniqueKey[index++] = '|';
          int length = thrownExceptionSignature.length;
          System.arraycopy(thrownExceptionSignature, 0, uniqueKey, index, length);
          index += length;
        } 
      } 
    } 
    return uniqueKey;
  }
  
  public final char[] constantPoolName() {
    return this.selector;
  }
  
  protected void fillInDefaultNonNullness(AbstractMethodDeclaration sourceMethod) {
    if (this.parameterNonNullness == null)
      this.parameterNonNullness = new Boolean[this.parameters.length]; 
    boolean added = false;
    int length = this.parameterNonNullness.length;
    for (int i = 0; i < length; i++) {
      if (!this.parameters[i].isBaseType())
        if (this.parameterNonNullness[i] == null) {
          added = true;
          this.parameterNonNullness[i] = Boolean.TRUE;
          if (sourceMethod != null)
            (sourceMethod.arguments[i]).binding.tagBits |= 0x100000000000000L; 
        } else if (sourceMethod != null && this.parameterNonNullness[i].booleanValue()) {
          sourceMethod.scope.problemReporter().nullAnnotationIsRedundant(sourceMethod, i);
        }  
    } 
    if (added)
      this.tagBits |= 0x400L; 
    if (this.returnType != null && 
      !this.returnType.isBaseType() && (
      this.tagBits & 0x180000000000000L) == 0L) {
      this.tagBits |= 0x100000000000000L;
    } else if (sourceMethod != null && (this.tagBits & 0x100000000000000L) != 0L) {
      sourceMethod.scope.problemReporter().nullAnnotationIsRedundant(sourceMethod, -1);
    } 
  }
  
  protected void fillInDefaultNonNullness18(AbstractMethodDeclaration sourceMethod, LookupEnvironment env) {
    if (hasNonNullDefaultFor(8, true)) {
      boolean added = false;
      int length = this.parameters.length;
      for (int i = 0; i < length; i++) {
        TypeBinding parameter = this.parameters[i];
        if (parameter.acceptsNonNullDefault()) {
          long existing = parameter.tagBits & 0x180000000000000L;
          if (existing == 0L) {
            added = true;
            if (!parameter.isBaseType()) {
              this.parameters[i] = env.createAnnotatedType(parameter, new AnnotationBinding[] { env.getNonNullAnnotation() });
              if (sourceMethod != null)
                (sourceMethod.arguments[i]).binding.type = this.parameters[i]; 
            } 
          } else if (sourceMethod != null && (parameter.tagBits & 0x100000000000000L) != 0L && 
            sourceMethod.arguments[i].hasNullTypeAnnotation(TypeReference.AnnotationPosition.MAIN_TYPE)) {
            sourceMethod.scope.problemReporter().nullAnnotationIsRedundant(sourceMethod, i);
          } 
        } 
      } 
      if (added)
        this.tagBits |= 0x400L; 
    } 
    if (this.returnType != null && hasNonNullDefaultFor(16, true) && this.returnType.acceptsNonNullDefault())
      if ((this.returnType.tagBits & 0x180000000000000L) == 0L) {
        this.returnType = env.createAnnotatedType(this.returnType, new AnnotationBinding[] { env.getNonNullAnnotation() });
      } else if (sourceMethod instanceof MethodDeclaration && (this.returnType.tagBits & 0x100000000000000L) != 0L && (
        (MethodDeclaration)sourceMethod).hasNullTypeAnnotation(TypeReference.AnnotationPosition.MAIN_TYPE)) {
        sourceMethod.scope.problemReporter().nullAnnotationIsRedundant(sourceMethod, -1);
      }  
  }
  
  public org.eclipse.jdt.internal.compiler.lookup.MethodBinding findOriginalInheritedMethod(org.eclipse.jdt.internal.compiler.lookup.MethodBinding inheritedMethod) {
    org.eclipse.jdt.internal.compiler.lookup.MethodBinding inheritedOriginal = inheritedMethod.original();
    TypeBinding superType = this.declaringClass.findSuperTypeOriginatingFrom((TypeBinding)inheritedOriginal.declaringClass);
    if (superType == null || !(superType instanceof ReferenceBinding))
      return null; 
    if (TypeBinding.notEquals((TypeBinding)inheritedOriginal.declaringClass, superType)) {
      org.eclipse.jdt.internal.compiler.lookup.MethodBinding[] superMethods = ((ReferenceBinding)superType).getMethods(inheritedOriginal.selector, inheritedOriginal.parameters.length);
      for (int m = 0, l = superMethods.length; m < l; m++) {
        if (superMethods[m].original() == inheritedOriginal)
          return superMethods[m]; 
      } 
    } 
    return inheritedOriginal;
  }
  
  public char[] genericSignature() {
    if ((this.modifiers & 0x40000000) == 0)
      return null; 
    StringBuffer sig = new StringBuffer(10);
    if (this.typeVariables != Binding.NO_TYPE_VARIABLES) {
      sig.append('<');
      for (int k = 0, m = this.typeVariables.length; k < m; k++)
        sig.append(this.typeVariables[k].genericSignature()); 
      sig.append('>');
    } 
    sig.append('(');
    int length;
    for (int i = 0; i < length; i++)
      sig.append(this.parameters[i].genericTypeSignature()); 
    sig.append(')');
    if (this.returnType != null)
      sig.append(this.returnType.genericTypeSignature()); 
    boolean needExceptionSignatures = false;
    length = this.thrownExceptions.length;
    int j;
    for (j = 0; j < length; j++) {
      if (((this.thrownExceptions[j]).modifiers & 0x40000000) != 0) {
        needExceptionSignatures = true;
        break;
      } 
    } 
    if (needExceptionSignatures)
      for (j = 0; j < length; j++) {
        sig.append('^');
        sig.append(this.thrownExceptions[j].genericTypeSignature());
      }  
    int sigLength = sig.length();
    char[] genericSignature = new char[sigLength];
    sig.getChars(0, sigLength, genericSignature, 0);
    return genericSignature;
  }
  
  public final int getAccessFlags() {
    return this.modifiers & 0x1FFFF;
  }
  
  public AnnotationBinding[] getAnnotations() {
    org.eclipse.jdt.internal.compiler.lookup.MethodBinding originalMethod = original();
    return originalMethod.declaringClass.retrieveAnnotations(originalMethod);
  }
  
  public long getAnnotationTagBits() {
    org.eclipse.jdt.internal.compiler.lookup.MethodBinding originalMethod = original();
    if ((originalMethod.tagBits & 0x200000000L) == 0L && originalMethod.declaringClass instanceof SourceTypeBinding) {
      ClassScope scope = ((SourceTypeBinding)originalMethod.declaringClass).scope;
      if (scope != null) {
        TypeDeclaration typeDecl = scope.referenceContext;
        AbstractMethodDeclaration methodDecl = typeDecl.declarationOf(originalMethod);
        if (methodDecl != null)
          ASTNode.resolveAnnotations((BlockScope)methodDecl.scope, methodDecl.annotations, originalMethod); 
        CompilerOptions options = scope.compilerOptions();
        if (options.isAnnotationBasedNullAnalysisEnabled) {
          boolean usesNullTypeAnnotations = scope.environment().usesNullTypeAnnotations();
          long nullDefaultBits = usesNullTypeAnnotations ? this.defaultNullness : (
            this.tagBits & 0x600000000000000L);
          if (nullDefaultBits != 0L && this.declaringClass instanceof SourceTypeBinding) {
            SourceTypeBinding declaringSourceType = (SourceTypeBinding)this.declaringClass;
            if (declaringSourceType.checkRedundantNullnessDefaultOne((ASTNode)methodDecl, methodDecl.annotations, nullDefaultBits, usesNullTypeAnnotations))
              declaringSourceType.checkRedundantNullnessDefaultRecurse((ASTNode)methodDecl, methodDecl.annotations, nullDefaultBits, usesNullTypeAnnotations); 
          } 
        } 
      } 
    } 
    return originalMethod.tagBits;
  }
  
  public Object getDefaultValue() {
    org.eclipse.jdt.internal.compiler.lookup.MethodBinding originalMethod = original();
    if ((originalMethod.tagBits & 0x800000000000000L) == 0L) {
      if (originalMethod.declaringClass instanceof SourceTypeBinding) {
        SourceTypeBinding sourceType = (SourceTypeBinding)originalMethod.declaringClass;
        if (sourceType.scope != null) {
          AbstractMethodDeclaration methodDeclaration = originalMethod.sourceMethod();
          if (methodDeclaration != null && methodDeclaration.isAnnotationMethod())
            methodDeclaration.resolve(sourceType.scope); 
        } 
      } 
      originalMethod.tagBits |= 0x800000000000000L;
    } 
    AnnotationHolder holder = originalMethod.declaringClass.retrieveAnnotationHolder(originalMethod, true);
    return (holder == null) ? null : holder.getDefaultValue();
  }
  
  public AnnotationBinding[][] getParameterAnnotations() {
    int length;
    if ((length = this.parameters.length) == 0)
      return null; 
    org.eclipse.jdt.internal.compiler.lookup.MethodBinding originalMethod = original();
    AnnotationHolder holder = originalMethod.declaringClass.retrieveAnnotationHolder(originalMethod, true);
    AnnotationBinding[][] allParameterAnnotations = (holder == null) ? null : holder.getParameterAnnotations();
    if (allParameterAnnotations == null && (this.tagBits & 0x400L) != 0L) {
      allParameterAnnotations = new AnnotationBinding[length][];
      if (this.declaringClass instanceof SourceTypeBinding) {
        SourceTypeBinding sourceType = (SourceTypeBinding)this.declaringClass;
        if (sourceType.scope != null) {
          AbstractMethodDeclaration methodDecl = sourceType.scope.referenceType().declarationOf(originalMethod);
          for (int i = 0; i < length; i++) {
            Argument argument = methodDecl.arguments[i];
            if (argument.annotations != null) {
              ASTNode.resolveAnnotations((BlockScope)methodDecl.scope, argument.annotations, (Binding)argument.binding);
              allParameterAnnotations[i] = argument.binding.getAnnotations();
            } else {
              allParameterAnnotations[i] = Binding.NO_ANNOTATIONS;
            } 
          } 
        } else {
          for (int i = 0; i < length; i++)
            allParameterAnnotations[i] = Binding.NO_ANNOTATIONS; 
        } 
      } else {
        for (int i = 0; i < length; i++)
          allParameterAnnotations[i] = Binding.NO_ANNOTATIONS; 
      } 
      setParameterAnnotations(allParameterAnnotations);
    } 
    return allParameterAnnotations;
  }
  
  public TypeVariableBinding getTypeVariable(char[] variableName) {
    for (int i = this.typeVariables.length; --i >= 0;) {
      if (CharOperation.equals((this.typeVariables[i]).sourceName, variableName))
        return this.typeVariables[i]; 
    } 
    return null;
  }
  
  public TypeVariableBinding[] getAllTypeVariables(boolean isDiamond) {
    TypeVariableBinding[] allTypeVariables = this.typeVariables;
    if (isDiamond) {
      TypeVariableBinding[] classTypeVariables = this.declaringClass.typeVariables();
      int l1 = allTypeVariables.length;
      int l2 = classTypeVariables.length;
      if (l1 == 0) {
        allTypeVariables = classTypeVariables;
      } else if (l2 != 0) {
        System.arraycopy(allTypeVariables, 0, allTypeVariables = new TypeVariableBinding[l1 + l2], 0, l1);
        System.arraycopy(classTypeVariables, 0, allTypeVariables, l1, l2);
      } 
    } 
    return allTypeVariables;
  }
  
  public boolean hasSubstitutedParameters() {
    return false;
  }
  
  public boolean hasSubstitutedReturnType() {
    return false;
  }
  
  public final boolean isAbstract() {
    return ((this.modifiers & 0x400) != 0);
  }
  
  public final boolean isBridge() {
    return ((this.modifiers & 0x40) != 0);
  }
  
  public final boolean isConstructor() {
    return (this.selector == TypeConstants.INIT);
  }
  
  public final boolean isDefault() {
    return (!isPublic() && !isProtected() && !isPrivate());
  }
  
  public final boolean isDefaultAbstract() {
    return ((this.modifiers & 0x80000) != 0);
  }
  
  public boolean isDefaultMethod() {
    return ((this.modifiers & 0x10000) != 0);
  }
  
  public final boolean isDeprecated() {
    return ((this.modifiers & 0x100000) != 0);
  }
  
  public final boolean isFinal() {
    return ((this.modifiers & 0x10) != 0);
  }
  
  public final boolean isImplementing() {
    return ((this.modifiers & 0x20000000) != 0);
  }
  
  public final boolean isMain() {
    if (this.selector.length == 4 && CharOperation.equals(this.selector, TypeConstants.MAIN) && (
      this.modifiers & 0x9) != 0 && 
      TypeBinding.VOID == this.returnType && 
      this.parameters.length == 1) {
      TypeBinding paramType = this.parameters[0];
      if (paramType.dimensions() == 1 && (paramType.leafComponentType()).id == 11)
        return true; 
    } 
    return false;
  }
  
  public final boolean isNative() {
    return ((this.modifiers & 0x100) != 0);
  }
  
  public final boolean isOverriding() {
    return ((this.modifiers & 0x10000000) != 0);
  }
  
  public final boolean isPrivate() {
    return ((this.modifiers & 0x2) != 0);
  }
  
  public final boolean isOrEnclosedByPrivateType() {
    if ((this.modifiers & 0x2) != 0)
      return true; 
    return (this.declaringClass != null && this.declaringClass.isOrEnclosedByPrivateType());
  }
  
  public final boolean isProtected() {
    return ((this.modifiers & 0x4) != 0);
  }
  
  public final boolean isPublic() {
    return ((this.modifiers & 0x1) != 0);
  }
  
  public final boolean isStatic() {
    return ((this.modifiers & 0x8) != 0);
  }
  
  public final boolean isStrictfp() {
    return ((this.modifiers & 0x800) != 0);
  }
  
  public final boolean isSynchronized() {
    return ((this.modifiers & 0x20) != 0);
  }
  
  public final boolean isSynthetic() {
    return ((this.modifiers & 0x1000) != 0);
  }
  
  public final boolean isUsed() {
    return ((this.modifiers & 0x8000000) != 0);
  }
  
  public boolean isVarargs() {
    return ((this.modifiers & 0x80) != 0);
  }
  
  public boolean isParameterizedGeneric() {
    return false;
  }
  
  public boolean isPolymorphic() {
    return false;
  }
  
  public final boolean isViewedAsDeprecated() {
    return ((this.modifiers & 0x300000) != 0);
  }
  
  public final int kind() {
    return 8;
  }
  
  public org.eclipse.jdt.internal.compiler.lookup.MethodBinding original() {
    return this;
  }
  
  public org.eclipse.jdt.internal.compiler.lookup.MethodBinding shallowOriginal() {
    return original();
  }
  
  public org.eclipse.jdt.internal.compiler.lookup.MethodBinding genericMethod() {
    return this;
  }
  
  public char[] readableName() {
    StringBuffer buffer = new StringBuffer(this.parameters.length + 20);
    if (isConstructor()) {
      buffer.append(this.declaringClass.sourceName());
    } else {
      buffer.append(this.selector);
    } 
    buffer.append('(');
    if (this.parameters != Binding.NO_PARAMETERS)
      for (int i = 0, length = this.parameters.length; i < length; i++) {
        if (i > 0)
          buffer.append(", "); 
        buffer.append(this.parameters[i].sourceName());
      }  
    buffer.append(')');
    return buffer.toString().toCharArray();
  }
  
  public final AnnotationBinding[] getTypeAnnotations() {
    return this.typeAnnotations;
  }
  
  public void setTypeAnnotations(AnnotationBinding[] annotations) {
    this.typeAnnotations = annotations;
  }
  
  public void setAnnotations(AnnotationBinding[] annotations) {
    this.declaringClass.storeAnnotations(this, annotations);
  }
  
  public void setAnnotations(AnnotationBinding[] annotations, AnnotationBinding[][] parameterAnnotations, Object defaultValue, LookupEnvironment optionalEnv) {
    this.declaringClass.storeAnnotationHolder(this, AnnotationHolder.storeAnnotations(annotations, parameterAnnotations, defaultValue, optionalEnv));
  }
  
  public void setDefaultValue(Object defaultValue) {
    org.eclipse.jdt.internal.compiler.lookup.MethodBinding originalMethod = original();
    originalMethod.tagBits |= 0x800000000000000L;
    AnnotationHolder holder = this.declaringClass.retrieveAnnotationHolder(this, false);
    if (holder == null) {
      setAnnotations(null, null, defaultValue, null);
    } else {
      setAnnotations(holder.getAnnotations(), holder.getParameterAnnotations(), defaultValue, null);
    } 
  }
  
  public void setParameterAnnotations(AnnotationBinding[][] parameterAnnotations) {
    AnnotationHolder holder = this.declaringClass.retrieveAnnotationHolder(this, false);
    if (holder == null) {
      setAnnotations(null, parameterAnnotations, null, null);
    } else {
      setAnnotations(holder.getAnnotations(), parameterAnnotations, holder.getDefaultValue(), null);
    } 
  }
  
  protected final void setSelector(char[] selector) {
    this.selector = selector;
    this.signature = null;
  }
  
  public char[] shortReadableName() {
    StringBuffer buffer = new StringBuffer(this.parameters.length + 20);
    if (isConstructor()) {
      buffer.append(this.declaringClass.shortReadableName());
    } else {
      buffer.append(this.selector);
    } 
    buffer.append('(');
    if (this.parameters != Binding.NO_PARAMETERS)
      for (int i = 0, length = this.parameters.length; i < length; i++) {
        if (i > 0)
          buffer.append(", "); 
        buffer.append(this.parameters[i].shortReadableName());
      }  
    buffer.append(')');
    int nameLength = buffer.length();
    char[] shortReadableName = new char[nameLength];
    buffer.getChars(0, nameLength, shortReadableName, 0);
    return shortReadableName;
  }
  
  public final char[] signature() {
    if (this.signature != null)
      return this.signature; 
    StringBuffer buffer = new StringBuffer(this.parameters.length + 20);
    buffer.append('(');
    TypeBinding[] targetParameters = this.parameters;
    boolean isConstructor = isConstructor();
    if (isConstructor && this.declaringClass.isEnum()) {
      buffer.append(ConstantPool.JavaLangStringSignature);
      buffer.append(TypeBinding.INT.signature());
    } 
    boolean needSynthetics = (isConstructor && this.declaringClass.isNestedType());
    if (needSynthetics) {
      ReferenceBinding[] syntheticArgumentTypes = this.declaringClass.syntheticEnclosingInstanceTypes();
      if (syntheticArgumentTypes != null)
        for (int i = 0, count = syntheticArgumentTypes.length; i < count; i++)
          buffer.append(syntheticArgumentTypes[i].signature());  
      if (this instanceof SyntheticMethodBinding)
        targetParameters = ((SyntheticMethodBinding)this).targetMethod.parameters; 
    } 
    if (targetParameters != Binding.NO_PARAMETERS)
      for (int i = 0; i < targetParameters.length; i++)
        buffer.append(targetParameters[i].signature());  
    if (needSynthetics) {
      SyntheticArgumentBinding[] syntheticOuterArguments = this.declaringClass.syntheticOuterLocalVariables();
      int count = (syntheticOuterArguments == null) ? 0 : syntheticOuterArguments.length;
      int i;
      for (i = 0; i < count; i++)
        buffer.append((syntheticOuterArguments[i]).type.signature()); 
      int extraLength;
      for (i = targetParameters.length, extraLength = this.parameters.length; i < extraLength; i++)
        buffer.append(this.parameters[i].signature()); 
    } 
    buffer.append(')');
    if (this.returnType != null)
      buffer.append(this.returnType.signature()); 
    int nameLength = buffer.length();
    this.signature = new char[nameLength];
    buffer.getChars(0, nameLength, this.signature, 0);
    return this.signature;
  }
  
  public final char[] signature(ClassFile classFile) {
    if (this.signature != null) {
      if ((this.tagBits & 0x800L) != 0L) {
        boolean bool1 = isConstructor();
        TypeBinding[] arrayOfTypeBinding = this.parameters;
        boolean bool2 = (bool1 && this.declaringClass.isNestedType());
        if (bool2) {
          ReferenceBinding[] syntheticArgumentTypes = this.declaringClass.syntheticEnclosingInstanceTypes();
          if (syntheticArgumentTypes != null)
            for (int i = 0, count = syntheticArgumentTypes.length; i < count; i++) {
              ReferenceBinding syntheticArgumentType = syntheticArgumentTypes[i];
              if ((syntheticArgumentType.tagBits & 0x800L) != 0L)
                Util.recordNestedType(classFile, (TypeBinding)syntheticArgumentType); 
            }  
          if (this instanceof SyntheticMethodBinding)
            arrayOfTypeBinding = ((SyntheticMethodBinding)this).targetMethod.parameters; 
        } 
        if (arrayOfTypeBinding != Binding.NO_PARAMETERS)
          for (int i = 0, max = arrayOfTypeBinding.length; i < max; i++) {
            TypeBinding targetParameter = arrayOfTypeBinding[i];
            TypeBinding leafTargetParameterType = targetParameter.leafComponentType();
            if ((leafTargetParameterType.tagBits & 0x800L) != 0L)
              Util.recordNestedType(classFile, leafTargetParameterType); 
          }  
        if (bool2)
          for (int i = arrayOfTypeBinding.length, extraLength = this.parameters.length; i < extraLength; i++) {
            TypeBinding parameter = this.parameters[i];
            TypeBinding leafParameterType = parameter.leafComponentType();
            if ((leafParameterType.tagBits & 0x800L) != 0L)
              Util.recordNestedType(classFile, leafParameterType); 
          }  
        if (this.returnType != null) {
          TypeBinding ret = this.returnType.leafComponentType();
          if ((ret.tagBits & 0x800L) != 0L)
            Util.recordNestedType(classFile, ret); 
        } 
      } 
      return this.signature;
    } 
    StringBuffer buffer = new StringBuffer((this.parameters.length + 1) * 20);
    buffer.append('(');
    TypeBinding[] targetParameters = this.parameters;
    boolean isConstructor = isConstructor();
    if (isConstructor && this.declaringClass.isEnum()) {
      buffer.append(ConstantPool.JavaLangStringSignature);
      buffer.append(TypeBinding.INT.signature());
    } 
    boolean needSynthetics = (isConstructor && this.declaringClass.isNestedType());
    if (needSynthetics) {
      ReferenceBinding[] syntheticArgumentTypes = this.declaringClass.syntheticEnclosingInstanceTypes();
      if (syntheticArgumentTypes != null)
        for (int i = 0, count = syntheticArgumentTypes.length; i < count; i++) {
          ReferenceBinding syntheticArgumentType = syntheticArgumentTypes[i];
          if ((syntheticArgumentType.tagBits & 0x800L) != 0L) {
            this.tagBits |= 0x800L;
            Util.recordNestedType(classFile, (TypeBinding)syntheticArgumentType);
          } 
          buffer.append(syntheticArgumentType.signature());
        }  
      if (this instanceof SyntheticMethodBinding)
        targetParameters = ((SyntheticMethodBinding)this).targetMethod.parameters; 
    } 
    if (targetParameters != Binding.NO_PARAMETERS)
      for (int i = 0, max = targetParameters.length; i < max; i++) {
        TypeBinding targetParameter = targetParameters[i];
        TypeBinding leafTargetParameterType = targetParameter.leafComponentType();
        if ((leafTargetParameterType.tagBits & 0x800L) != 0L) {
          this.tagBits |= 0x800L;
          Util.recordNestedType(classFile, leafTargetParameterType);
        } 
        buffer.append(targetParameter.signature());
      }  
    if (needSynthetics) {
      SyntheticArgumentBinding[] syntheticOuterArguments = this.declaringClass.syntheticOuterLocalVariables();
      int count = (syntheticOuterArguments == null) ? 0 : syntheticOuterArguments.length;
      int i;
      for (i = 0; i < count; i++)
        buffer.append((syntheticOuterArguments[i]).type.signature()); 
      int extraLength;
      for (i = targetParameters.length, extraLength = this.parameters.length; i < extraLength; i++) {
        TypeBinding parameter = this.parameters[i];
        TypeBinding leafParameterType = parameter.leafComponentType();
        if ((leafParameterType.tagBits & 0x800L) != 0L) {
          this.tagBits |= 0x800L;
          Util.recordNestedType(classFile, leafParameterType);
        } 
        buffer.append(parameter.signature());
      } 
    } 
    buffer.append(')');
    if (this.returnType != null) {
      TypeBinding ret = this.returnType.leafComponentType();
      if ((ret.tagBits & 0x800L) != 0L) {
        this.tagBits |= 0x800L;
        Util.recordNestedType(classFile, ret);
      } 
      buffer.append(this.returnType.signature());
    } 
    int nameLength = buffer.length();
    this.signature = new char[nameLength];
    buffer.getChars(0, nameLength, this.signature, 0);
    return this.signature;
  }
  
  public final int sourceEnd() {
    AbstractMethodDeclaration method = sourceMethod();
    if (method == null) {
      if (this.declaringClass instanceof SourceTypeBinding)
        return ((SourceTypeBinding)this.declaringClass).sourceEnd(); 
      return 0;
    } 
    return method.sourceEnd;
  }
  
  public AbstractMethodDeclaration sourceMethod() {
    SourceTypeBinding sourceType;
    if (isSynthetic())
      return null; 
    try {
      sourceType = (SourceTypeBinding)this.declaringClass;
    } catch (ClassCastException classCastException) {
      return null;
    } 
    AbstractMethodDeclaration[] methods = (sourceType.scope != null) ? sourceType.scope.referenceContext.methods : null;
    if (methods != null)
      for (int i = methods.length; --i >= 0;) {
        if (this == (methods[i]).binding)
          return methods[i]; 
      }  
    return null;
  }
  
  public LambdaExpression sourceLambda() {
    return null;
  }
  
  public final int sourceStart() {
    AbstractMethodDeclaration method = sourceMethod();
    if (method == null) {
      if (this.declaringClass instanceof SourceTypeBinding)
        return ((SourceTypeBinding)this.declaringClass).sourceStart(); 
      return 0;
    } 
    return method.sourceStart;
  }
  
  public org.eclipse.jdt.internal.compiler.lookup.MethodBinding tiebreakMethod() {
    return this;
  }
  
  public String toString() {
    StringBuffer output = new StringBuffer(10);
    if ((this.modifiers & 0x2000000) != 0)
      output.append("[unresolved] "); 
    ASTNode.printModifiers(this.modifiers, output);
    output.append((this.returnType != null) ? this.returnType.debugName() : "<no type>");
    output.append(" ");
    output.append((this.selector != null) ? new String(this.selector) : "<no selector>");
    output.append("(");
    if (this.parameters != null) {
      if (this.parameters != Binding.NO_PARAMETERS)
        for (int i = 0, length = this.parameters.length; i < length; i++) {
          if (i > 0)
            output.append(", "); 
          output.append((this.parameters[i] != null) ? this.parameters[i].debugName() : "<no argument type>");
        }  
    } else {
      output.append("<no argument types>");
    } 
    output.append(") ");
    if (this.thrownExceptions != null) {
      if (this.thrownExceptions != Binding.NO_EXCEPTIONS) {
        output.append("throws ");
        for (int i = 0, length = this.thrownExceptions.length; i < length; i++) {
          if (i > 0)
            output.append(", "); 
          output.append((this.thrownExceptions[i] != null) ? this.thrownExceptions[i].debugName() : "<no exception type>");
        } 
      } 
    } else {
      output.append("<no exception types>");
    } 
    return output.toString();
  }
  
  public TypeVariableBinding[] typeVariables() {
    return this.typeVariables;
  }
  
  public boolean hasNonNullDefaultFor(int location, boolean useTypeAnnotations) {
    if ((this.modifiers & 0x4000000) != 0)
      return false; 
    if (useTypeAnnotations) {
      if (this.defaultNullness != 0)
        return ((this.defaultNullness & location) != 0); 
    } else {
      if ((this.tagBits & 0x200000000000000L) != 0L)
        return true; 
      if ((this.tagBits & 0x400000000000000L) != 0L)
        return false; 
    } 
    return this.declaringClass.hasNonNullDefaultFor(location, useTypeAnnotations);
  }
  
  public boolean redeclaresPublicObjectMethod(Scope scope) {
    ReferenceBinding javaLangObject = scope.getJavaLangObject();
    org.eclipse.jdt.internal.compiler.lookup.MethodBinding[] methods = javaLangObject.getMethods(this.selector);
    for (int i = 0, length = (methods == null) ? 0 : methods.length; i < length; i++) {
      org.eclipse.jdt.internal.compiler.lookup.MethodBinding method = methods[i];
      if (method.isPublic() && !method.isStatic() && method.parameters.length == this.parameters.length)
        if (MethodVerifier.doesMethodOverride(this, method, scope.environment()))
          return true;  
    } 
    return false;
  }
  
  public boolean isVoidMethod() {
    return (this.returnType == TypeBinding.VOID);
  }
}
